#!/usr/bin/env python

import sys
import re

keywords = ["if", "then", "do", "else", "elseif", "variable",
            "switch", "case", "exitscript", "exitreturning",
            "for", "while", "return", "break", "continue"]
toplevel = ["definescript", "defineconstant", "globalvariable",
            "script", "plotscript", "subscript", "include"]
moreops = ["and", "or", "xor", "not", "setvariable", "increment",
           "decrement", "mod", "sqrt", "abs", "sign"]
flow_continuations = ["do", "then", "else"]#, "elseif"]

ops = [
    "\+ *=" ,
    "- *=" ,
    "\$ *\+" ,
    "\$ *=" ,
    "\+" ,
    "- *-" ,
    "/" ,
    "\*" ,
    "\^ *\^" ,
    "\^" ,
    "= *=" ,
    "< *>" ,
    "> *>" ,
    "< *<" ,
    "< *=" ,
    "> *=" ,
    ">" ,
    "<" ,
    ": *=" ,
    "& *&" ,
    "\| *\|",
]

# Token types
definitions = [
#   ("sep", "[,\+\*/^|&$]+"),
    ("comment", "#"),
    ("sep", ",+"),
    ("open", "\("),
    ("close", "\)"),
    ("op", "(" + ")|(".join(ops) + ")"),
    ("string", r'"([^"\\]+|(\\.))*"'),
    ("badstring", r'".*'),
#   ("text", "[a-z0-9 :-]+")
#other token types:
#   "text"
#   "error"
#   "newline"
#   "keyword"
#TODO   "whitespace"
]


# from http://www.gooli.org/blog/a-simple-lexer-in-python/
class Lexer(object):
    def __init__(self, definitions):
        self.definitions = definitions
        
        # create a compound regular expression from the lexer definitions
        parts = []
        for name, part in definitions:
#           for i in xrange(len(definitions)):
            parts.append("(?P<%s>%s)" % (name, part))
#           parts.append("(?P<%s>%s)" % (i, definitions[i][1]))
            self.regexpString = "|".join(parts)
            self.regexp = re.compile(self.regexpString, re.MULTILINE | re.IGNORECASE)
    
    def parse(self, text):
        # yield lexemes
        for match in self.regexp.finditer(text):
            found = False
            for name, rexp in self.definitions:
                m = match.group(name)
                if m is not None:
                    yield (name, m, match)
                   #yield (defintions[name][0], m, match)
                    break


def norm(string):
    return string.replace(" ", "").strip().lower()

class Colouring:
    colours = {}
    def __init__(self, style):
        if style == "simple":
            self.colours = {"comment": "#404040", "op": "#4070c0", "open": "#00b000", "close": "#30a030", "error": "#ff4040", "keyword": "#0000a0", "string": "#a88000"}
        else:
            self.colours = {"error": "#ff4040"}
        self.colours["badstring"] = self.colours["error"]
    
    def write(self, text, type):
        if type in self.colours:
            if text in (")", "(") and type == "error":  # these are hard to see
                text += "###"
            return "<font color=\"" + self.colours[type] + "\">" + text + "</font>"
        else:
            return text

class Token:
    def __init__(self, type='', value='', depth=0):
        self.type = type
        self.value = value
        self.norm = norm(value)
        self.lineno = fileinput.lineno()
        self.args = 0
        self.depth = depth
        self.parent = None  #used by close and text types only
        self.bad = False  # If true, display in "error" font
#       global parser
#       if len(parser.parent_stack):
#           self.parentflow = parser.parent_stack[-1]  #lowest enclosing do,then or else flow control
#       else:
#           self.parentflow = None
        self.foldargs = False  #should only affect do, then, else



class FormatOptions:
    spaceflow = False #space after if, then, switch, do, etc

class Parser:
    depth = 0
    #linenum
    lasttok = ''
    parent_stack = []
    result = ''
    localvars = set()
    globalvars = set()


    class Formatter:
        queue = []
        def __init__(self, parent):
            self.up = parent

        def out(self, token):
#           print "depth = ", token.depth, " '" + token.value + "'"
#           repr (token)
            self.queue.append(token)
            if len(self.up.parent_stack) and self.up.parent_stack[-1].args > 1:
                for flow in self.up.parent_stack:
                    flow.foldargs = False


        def flush(self):
            write = self.up.colouring.write
            newl = True
            currentflow = None
            depthdebt = 0
            while len(self.queue):
                token = self.queue.pop(0)
                if len(self.queue):
                    next = self.queue[0]
                else:
                    next = None

                #ditch whitespace (TODO: keep in some cases!)
                if token.type == "text" and token.value.strip() == "":
                    continue

                if token.norm in flow_continuations:
                    currentflow = token
                #just in case
                if token.parent and token.parent.norm in flow_continuations:
                    currentflow = token.parent

                #ditch newlines when lumping
                if token.type == "newline" and token.parent and token.parent.foldargs: #currentflow and currentflow.foldargs:
#                   print "ditching newline!!"
                    continue


                #insert newline if after do/then/else and not folding   (????)
                if token.type == "open" and currentflow and not currentflow.foldargs and token.lineno == currentflow.lineno and next and next.type != "newline":
                    #do a newline next
#                   self.queue.insert(0, token)
                    self.queue.insert(0, Token("newline", "\n"))
                    next = self.queue[0]
    #               continue


            #   if token.type == "close" or token.type == "open:
            #       token.value = token.value.strip()

                # actually print!!
                if newl:
                    newl = False
                    token.value = token.value.lstrip(" ").lstrip("\t")

                    depth = token.depth + depthdebt
                    if depth >= 0:
                        self.up.result += '\t' * depth
                    else:
                        self.up.result += write("# TOO MANY ENDS" * -depth + "\n", "error")
                        depthdebt -= token.depth


                if token.norm in keywords:
                    token.type = "keyword"
                if token.norm in moreops:
                    token.type = "op"

                colour_class = token.type
                if token.bad:
                    colour_class = "error"
                self.up.result += write(token.value, colour_class)   # + "<<args=" + str(token.args) + ">>"

                if token.type == "newline" or token.type == "comment":
                    newl = True
                # end printing!!


                if token.type == "close" and token.parent:
                    if token.parent.norm in ['do', 'then', 'else', 'elseif', 'case'] and next and next.type != "newline":
#                       print "<<adding newline after ", token.value, " parent", token.parent.value, ">>"
                        self.queue.insert(0, Token("newline", "\n"))


            self.queue = None

    def __init__(self, lexer, colouring):
        self.lex = lexer
        self.colouring = colouring
        self.output = self.Formatter(self)


#   def print_tok(self, line, type):
    
    def process_token(self, t):  #token
        write = self.output.out
        if t.type == "text":
            if t.norm == "begin":
                t.type = "open"
            elif t.norm == "end":
                t.type = "close"
#       print t.type + " '" + t.value + "'"
    
        if t.type == "open":
            self.depth += 1
#           self.parent_stack.append( (self.lasttok, fileinput.lineno(), 0) )
            self.parent_stack.append( self.lasttok )
        #   if lasttok in flow_continuations:
                
            self.lasttok = Token()
        elif t.type == "close":
            self.depth -= 1
            if self.depth >= 0:
#               print "closing " + self.parent_stack.pop()
                t.parent = self.parent_stack.pop()
        elif t.type == "text" and t.norm != "":   #whitespace doesn't count!
            if len(self.parent_stack) > 0:
#               print t.value + " arg of " + self.parent_stack[-1].value
                self.parent_stack[-1].args += 1

                if self.parent_stack[-1].lineno == t.lineno:
                    self.parent_stack[-1].foldargs = True

                if self.parent_stack[-1].norm == "variable":
                    self.localvars.add(t.norm)

                if self.parent_stack[-1].norm == "globalvariable":
                    self.globalvars.add(t.norm)

                if self.parent_stack[-1].norm == "script" or self.parent_stack[-1].norm == "plotscript":
                    self.localvars.clear()
            self.lasttok = t

        if t.norm == "case":
            if self.check_stack("__caseblock"):
                self.depth -= 1
                self.parent_stack.pop()
            elif not self.check_stack("do"):#, "switch"):
                t.bad = True

        # After a case parented to a switch-do, move inside a phantom __caseblock
        # (after that, switch is no longer the parent of the remaining tokens, so
        # only happens once)
        if t.type == "close" and t.parent.norm == "case":
            if self.check_stack("do"):#, "switch"):
                self.depth += 1
                self.parent_stack.append( Token("text", "__caseblock", self.depth) )
            else:
                t.bad = True
        if t.type == "close" and t.parent.norm == "__caseblock":
            self.depth -= 1
            assert self.parent_stack[-1].norm == "do"
            t.parent = self.parent_stack.pop()

        t.depth = self.depth
        write(t)

    def check_stack(self, *tokens):
        """Check the final tokens on the stack (reading from last backwards) are as given.
        Arguments is a list of strings (token norms)"""
        if len(self.parent_stack) < len(tokens):
            return False
        for idx, tok in enumerate(tokens):
            if self.parent_stack[-1 - idx].norm != tok:
                return False
        return True

    def process_line(self, line):
        write = self.output.out
        
    #   script = script.lstrip()
        textstart = 0
        for tokenType, tokenValue, m in self.lex.parse(line):
            token = Token(tokenType, tokenValue)
#           print token.value
    #       print m.groups()
            
            
            if m.start() > textstart:
                self.process_token( Token("text", line[textstart:m.start()], self.depth) )
            textstart = m.end()
            
            self.process_token(token)
                    
            if tokenType == "comment":
                # comments are meant to imply newlines, but apparently not working
                write( Token("comment", line[m.start():]) )
                write(Token("newline", "\n"))
                return
            
#           if tokenType != "sep":
#               lastnonsep = m.end(0)

        if len(line) > textstart:
            self.process_token( Token("text", line[textstart:], self.depth) )

        write(Token("newline", "\n"))

    def flush(self):
        self.output.flush()
        print self.result


#fsock = open('error.log', 'w')
#sys.stderr = fsock

style = 'none'
cstyle = 'simple'
#cstyle = 'none'
if len(sys.argv) >= 2:
    style = sys.argv[1]
if len(sys.argv) >= 3:
    cstyle = sys.argv[2]

parse = Parser(Lexer(definitions), Colouring(cstyle))

print "<pre><b>"

import fileinput
for line in fileinput.input('-'):
    parse.process_line(line.strip())

parse.flush()

print "</b></pre>"

#print "bracket balance (should be 0) =", parse.depth, "
#print "syntax highlighting style =", cstyle


#lines = [x.lstrip() for x in sys.argv[1].splitlines()]
#for s in lines:
#   outputting(s)

